/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
#include "autoconfig.h"
#endif

#include <digitalSTROM/ds485-client-interface.h>
#include <autosprintf.h>

#include "json_dsm_web_handler.h"
#include "ds-assistant.h"
#include "exceptions.h"
#include "storage.h"
#include "database.h"
#include "hwstats.h"
#include "log.h"
#include "gettext_helper.h"

using gnu::autosprintf;

namespace dSA
{

JSONdSMWebHandler::JSONdSMWebHandler(struct mg_connection* connection,
                               const struct mg_request_info* request_info) :
                               JSONWebHandler(connection, request_info) {}

boost::shared_ptr<const dSM> JSONdSMWebHandler::getDSMfromParam()
{
    std::string dsid = m_parameters["dsid"];
    if (dsid.empty())
    {
        throw _Exception(_("missing dSM dSID parameter"));
    }

    dsid_t dsm_dsid = util::str2dsid(dsid);
    boost::shared_ptr<dSM> dsm = Storage::getInstance()->getDSM(dsm_dsid);
    return dsm;
}

std::string JSONdSMWebHandler::restartRegistration()
{
    std::ostringstream result;

    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    dS_Assistant::getInstance()->dS485()->restartRegistration(dsm->getDSID());

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdSMWebHandler::getCurrentTemperature()
{
    std::ostringstream result;

    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    uint16_t t;
    dS_Assistant::getInstance()->dS485()->getDSMtemperature(dsm->getDSID(), &t);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    json_object_object_add(res, "temperature", json_object_new_int(t));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdSMWebHandler::blink()
{
    std::ostringstream result;

    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    uint8_t flags;
    dS_Assistant::getInstance()->dS485()->getDisplayMode(dsm->getDSID(), &flags);
    dS_Assistant::getInstance()->dS485()->setDisplayMode(dsm->getDSID(), 0x41);
    sleep(4);
    dS_Assistant::getInstance()->dS485()->setDisplayMode(dsm->getDSID(), flags);

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdSMWebHandler::getStatistic()
{
    std::ostringstream result;

    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    uint8_t length;
    uint8_t index;
    uint8_t data[256];
    try
    {
        index = 0;
        dS_Assistant::getInstance()->dS485()->getStatisticUpstream(
                                    dsm->getDSID(), 0, &length, data + index);
        index += length;
        dS_Assistant::getInstance()->dS485()->getStatisticUpstream(
                                    dsm->getDSID(), 1, &length, data + index);
        index += length;
    }
    catch (Exception& ex)
    {
        FILE_LOG(logERROR) << "[Task Meter Statistic] Query from dsm " <<
                util::str(dsm->getDSID()) << " failed: " << ex.getMessage();
        return ex.getMessage();
    }

    uint16_t sversion = *(uint16_t *) data;
    boost::shared_ptr<StatisticMeter> pstats(new StatisticMeter(sversion));

    for (int i = 0; i < index; i++)
    {
        pstats->pushValue(data[i], 0);
    }
    pstats->parseValues();

    Database* db = Database::getInstance();
    db->saveStatisticMeter(0, dsm->getDSID(), pstats->m_value_map);

    DSID_PREFIX_LOG(logINFO, dsm->getDSID(), "stat_")
        << "Statistic Query Results:";
    boost::shared_ptr<std::map<statistic_field_t, std::string > > sresults =
        pstats->formatValues();
    for (std::map<statistic_field_t, std::string>::iterator it = sresults->begin();
         it != sresults->end();
         it++)
    {
        std::pair<std::string,std::string> desc = db->getStatisticDescription((*it).first);
        DSID_PREFIX_LOG(logINFO, dsm->getDSID(), "stat_") << desc.first << " = " << (*it).second;
        (*it).second = desc.first;
    }


    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    for (std::list<boost::shared_ptr<StatisticValue> >::iterator it = pstats->m_value_map->begin();
         it != pstats->m_value_map->end();
         it++)
    {
        json_object *vres = json_object_new_object();
        json_object_object_add(vres, "statid", json_object_new_int((*it)->m_field));
        json_object_object_add(vres, "value", json_object_new_int((*it)->m_value));
        json_object_object_add(res, sresults->at((*it)->m_field).c_str(), vres);
    }
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdSMWebHandler::resetStatistic()
{
    std::ostringstream result;

    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    dS_Assistant::getInstance()->dS485()->resetStatisticUpstream(
            dsm->getDSID());
    DSID_PREFIX_LOG(logINFO, dsm->getDSID(), "stat_") << "RESET";

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdSMWebHandler::getLedInfo()
{
    std::ostringstream result;
    boost::shared_ptr<std::list<dsid_t> > dsm_list;
    std::string dsid = m_parameters["dsid"];
    Storage *storage = Storage::getInstance();

    if (dsid.empty())
    {
        dsm_list = storage->getDSMList();
    }
    else
    {
        dsid_t dsm_dsid = util::str2dsid(dsid);
        dsm_list = boost::shared_ptr<std::list<dsid_t> >(new std::list<dsid_t>);
        dsm_list->push_back(dsm_dsid);
    }

    json_object *ret = json_ok();
    json_object *data = json_object_new_object();
    json_object *dsms = json_object_new_array();
    json_object_object_add(data, "dsms", dsms);

    while (!dsm_list->empty())
    {
        uint8_t hardware = 0;
        uint8_t DSP = 0;
        uint8_t temperature = 0;
        uint8_t config = 0;
        uint8_t dS485 = 0;
        uint8_t line = 0;
        std::string status(_("Offline"));
        dSM::led_info code = dSM::LED_off;

        boost::shared_ptr<const dSM> dsm = storage->getDSM(dsm_list->front());
        dsm_list->pop_front();

        if (dsm == NULL)
        {
            continue;
        }

        json_object *jdsm = json_object_new_object();
        json_object_object_add(jdsm, "dsid",
                json_object_new_string(util::str(dsm->getDSID()).c_str()));
        json_object *colors = json_object_new_array();

        if (dsm->getPresent())
        {
            try
            {
                Database::getInstance()->getDSMState(dsm,
                        &hardware, &DSP, &temperature, &config, &dS485, &line);
                code = dSM::LED_green;
            }
            catch (Exception &ex)
            {
                if (ex.getErrorCode() == ERROR_RESPONSE_TIMEOUT)
                {
                    status = _("Timeout");
                }
                else
                {
                    status = _("Error") + std::string(": ") + ex.getMessage();
                }
            }
        }

        // add status text, highest priority first
        if (temperature > 0)
        {
            code = dSM::LED_red_yellow;
            switch (temperature)
            {
                case 1:
                    status += _("Over temperature - communication deactivated");
                    break;
                case 2:
                    status += _("Over temperature - dSM must be restarted");
                    break;
            }
        }
        if ((hardware == 2) || (DSP == 1))
        {
            code = dSM::LED_red_black;
            if (status.length())
            {
                status += ", ";
            }
            status += _("Hardware failure");
        }
        if (hardware == 3)
        {
            code = dSM::LED_red_black;
            if (status.length())
            {
                status += ", ";
            }
            status += _("Flash failure");
        }
        if (hardware == 1)
        {
            code = dSM::LED_red_blue;
            if (status.length())
            {
                status += ", ";
            }
            status += _("Updating firmware");
        }
        if (config > 0)
        {
            code = dSM::LED_magenta;
            if (status.length())
            {
                status += ", ";
            }
            status += _("Invalid configuration - dSM must be replaced");
        }

        // finally
        if (code == dSM::LED_green)
        {
            status = _("Ready");
            if (line > 0)
            {
                status += std::string(", ") + _("Device registration running");
            }
        }

        switch (code)
        {
            case dSM::LED_green:
                json_object_array_add(colors, json_object_new_string("green"));
                break;
            case dSM::LED_red_black:
                json_object_array_add(colors, json_object_new_string("red"));
                json_object_array_add(colors, json_object_new_string("black"));
                break;
            case dSM::LED_red_yellow:
                json_object_array_add(colors, json_object_new_string("red"));
                json_object_array_add(colors, json_object_new_string("yellow"));
                break;
            case dSM::LED_red_blue:
                json_object_array_add(colors, json_object_new_string("red"));
                json_object_array_add(colors, json_object_new_string("blue"));
                break;
            case dSM::LED_magenta:
                json_object_array_add(colors,
                        json_object_new_string("magenta"));
                break;
            case dSM::LED_off:
                json_object_array_add(colors, json_object_new_string("none"));
                break;
            default:
                break;
        }
        json_object_object_add(jdsm, "ledcode", json_object_new_int(code));
        json_object_object_add(jdsm, "ledcolors", colors);
        json_object_object_add(jdsm, "status", json_object_new_string(status.c_str()));

        json_object_array_add(dsms, jdsm);
    }

    json_object_object_add(ret, "data", data);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();

}

std::string JSONdSMWebHandler::getDeviceRegistrationMode()
{
    std::ostringstream result;
    uint8_t f;

    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    dS_Assistant::getInstance()->dS485()->getPropertyFlags(dsm->getDSID(), &f);

    json_object *ret = json_ok();
    json_object *res = json_object_new_object();

    // Bit 4: Register of new devices disabled
    json_object_object_add(res, "actionsfromnewdevicesenabled",
                           json_object_new_boolean(!(f & (1 << 4))));
    json_object_object_add(ret, "data", res);

    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

std::string JSONdSMWebHandler::setDeviceRegistrationMode()
{
    std::ostringstream result;
    boost::shared_ptr<const dSM> dsm = getDSMfromParam();

    std::string reg = m_parameters["ignoreActionsFromNewDevices"];
    if (reg.empty())
    {
        throw _Exception(_("parameter ignoreActionsFromNewDevices missing"));
    }
    if ((reg != "yes") && (reg != "no"))
    {
        throw _Exception(_("invalid ignoreActionsFromNewDevices parameter value"));
    }

    uint8_t f;
    dS_Assistant::getInstance()->dS485()->getPropertyFlags(dsm->getDSID(), &f);
    // Bit 4: Register of new devices disabled
    if (reg == "yes")
    {
        f |= (1 << 4);
    }
    else
    {
        f &= ~(1 << 4);
    }
    dS_Assistant::getInstance()->dS485()->setPropertyFlags(dsm->getDSID(), f);

    json_object *ret = json_ok();
    result << json_object_to_json_string(ret);
    json_object_put(ret);
    return result.str();
}

void JSONdSMWebHandler::process()
{
    try
    {
        std::string result;
        std::string header = http_ok();

        if (m_command == "restartRegistration")
        {
            result = restartRegistration();
        }
        else if (m_command == "getCurrentTemperature")
        {
            result = getCurrentTemperature();
        }
        else if (m_command == "blink")
        {
            result = blink();
        }
        else if (m_command == "getStatistic")
        {
            result = getStatistic();
        }
        else if (m_command == "resetStatistic")
        {
            result = resetStatistic();
        }
        else if (m_command == "getLedInfo")
        {
            result = getLedInfo();
        }
        else if (m_command == "getDeviceRegistrationMode")
        {
            result = getDeviceRegistrationMode();
        }
        else if (m_command == "setDeviceRegistrationMode")
        {
            result = setDeviceRegistrationMode();
        }
        else
        {
            failure("unknown function called");
            return;
        }

        mg_write(m_connection, header.c_str(), header.length());
        mg_write(m_connection, result.c_str(), result.length());
    }
    catch (Exception &ex)
    {
        failure(ex.getMessage().c_str());
    }
    catch (...)
    {
        failure(autosprintf(_("exception raised while processing call %s"),
                            m_command.c_str()));
    }
}

} // namespace

